/* =============================================================================
 UPC - EETAC - CSD - http://digsys.upc.edu
 P10: An example on how organise a project as a FSM

 * Basic ideas on the design of a stepper motor controller 
Features:
  Specifications https://digsys.upc.edu/csd/units/exams/EX2/2021Q1_CSD_EXAM2.pdf
 
 * Programming target chip PIC18F4520. 
 * C language, coding in FSM - CSD style
 * External stepping CLK. 
 * Inhibit or run modes depending on var_INH switch 
 * Clockwise or counterclockwise selected by var_CW switch
 * Four coils for unipolar full-step mode
 * Four diodes for representing the driving sequence 

Plan:  Study problem planning at
    https://digsys.upc.edu/csd/units/exams/EX2/2021Q1_CSD_EXAM2_sol_P3.pdf  
 
- Step 1: Prepare the hardware Proteus project. For instance:
    L4 - RB5; L3 - RA3, L2 - RA2, L1 - RA1
    Q4 - RB4; Q3 - RB3, Q2 - RB2, Q1 - RB1
    
    var_CW - RB7, var_INH - RB6
    External CLK is RB0/INT0
  		
- Step 2: Software. Translate planning ideas and flowcharts into C code.
    Compile, run, debug using Proteus     

- Step 3: Prototype the experiment for instance in a PICDEM2 Plus board     
  

NOTE:The C code exemplified here is not optimised in any sense. The CSD style of 
programming aims clarity and facilitating the comprehension of the basic ideas 
on microcontrollers taking advantage of the previous content on combinational
and sequential systems in CSD. 
============================================================================= */
#include <xc.h>	// This header file includes the device specific headers 
                // like <pic18f4520.h>
                // This file contains all the microcontroller declarations


/* 	============================================================================
        Definitions
        ===================================================================== */

// This is the state enumeration. 
// Each state label is "coded" simply using a capital letter for easy debugging		
#define 	State_9	 		'A' 
#define 	State_5			'B'
#define 	State_6 		'C'
#define 	State_10		'D'

/* 	============================================================================
        Function prototypes fr organising the FSM
    ========================================================================= */
void init_system(void);
char output_logic(void);
char state_logic(void);

void read_inputs(void); 
void write_outputs(void);

// This is the interrupt service routine
static void interrupt ISR(void); 


/* 	============================================================================
        Global variables
============================================================================= */
	// let's use byte (char) types to make it easy the debugging
static char var_CW;             // Inputs polled
static char var_INH;            // Inputs polled

static char var_Q;            // Output LED
static char var_L;            // Output coils

static char var_current_state;      // state variable
// The state variable which will be updated ("asynchronously") every time that
// "state_logic()" is executed. Realise that indeed, only after having an 
// interrupt INT0 the system will take a new decision where to go.

static char var_CLK_flag;       // The pulse train input detected by interrupts


/* 	============================================================================
        Main function
        ======================================================================*/
void main(void){
    init_system ();
    while(1) {                      /* loop forever the FSM */
        if (var_CLK_flag == 1){ 
            // it means that there is an interrupt and a new STEP is possible
            read_inputs(); // Poll switch values    		   
            if (state_logic()) break;
            if (output_logic()) break;
            write_outputs();
            
            var_CLK_flag = 0;	// Clear the pulse detection flag
        }        
    }
    while (1){
        __nop();// The system has crashed in an illegal state and needs 
                //attention. This loop is never reached in normal operation. 
                // The only way to scape it is clicking the CD
    }   
}
	

/* 	============================================================================
        Function definitions
        ===================================================================== */

//*****************************************************************************
//Init operations  
//*****************************************************************************
void init_system(void) {
// Microcontroller general configuration bits. No need to change them.
// PIC18F4520 Configuration Bit Settings. They can be generated automatically 
// in the MPLAB X IDE

// CONFIG1H
#pragma config OSC = XT       // Oscillator Section bits (XT oscillator)
#pragma config FCMEN = OFF     
// Fail-Safe Clock Monitor E bit (Fail-Safe Clock Monitor disabled)
#pragma config IESO = OFF       
// Internal/External Osc Switch-over bit (Oscillator Switch-over mode disabled)

// CONFIG2L
#pragma config PWRT = ON       // Power-up Timer E bit (PWRT enabled)
#pragma config BOREN = SBORDIS 
// Brown-out Reset E bits (Brown-out Reset enabled in hardware 
// only (SBOREN is disabled))
#pragma config BORV = 3    // Brown Out Reset Voltage bits (Minimum setting)

// CONFIG2H
#pragma config WDT = OFF
// Watchdog Timer E bit (WDT disabled (control is on the SWDTEN bit))
#pragma config WDTPS = 32768    
// Watchdog Timer Postscale Sect bits (1:32768)

// CONFIG3H
#pragma config CCP2MX = PORTC  
// CCP2 MUX bit (CCP2 input/output is multiplexed with RC1)
#pragma config PBADEN = OFF    
// PORTB A/D E bit (PORTB<4:0> pins are configured 
// as digital I/O on Reset)
#pragma config LPT1OSC = OFF   
// Low-Power Timer1 Oscillator E bit (Timer1 configured for 
// higher power operation)
#pragma config MCLRE = ON
// MCLR Pin E bit (MCLR pin enabled; RE3 input pin disabled)
 
// CONFIG4L
#pragma config STVREN = ON      
// Stack Full/Underflow Reset E bit (Stack full/underflow will cause Reset)
#pragma config LVP = OFF
// Single-Supply ICSP E bit (Single-Supply ICSP disabled)
#pragma config XINST = OFF
// Extended Instruction Set E bit (Instruction set extension and Indexed 
// Addressing mode disabled (Legacy mode))

// CONFIG5L
#pragma config CP0 = OFF
// Code Protection bit (Block 0 (000800-001FFFh) not code-protected)
#pragma config CP1 = OFF        
// Code Protection bit (Block 1 (002000-003FFFh) not code-protected)
#pragma config CP2 = OFF        
// Code Protection bit (Block 2 (004000-005FFFh) not code-protected)
#pragma config CP3 = OFF        
// Code Protection bit (Block 3 (006000-007FFFh) not code-protected)

// CONFIG5H
#pragma config CPB = OFF        
// Boot Block Code Protection bit 
// (Boot block (000000-0007FFh) not code-protected)
#pragma config CPD = OFF        
// Data EEPROM Code Protection bit (Data EEPROM not code-protected)

// CONFIG6L
#pragma config WRT0 = OFF       
// Write Protection bit (Block 0 (000800-001FFFh) not write-protected)
#pragma config WRT1 = OFF       
// Write Protection bit (Block 1 (002000-003FFFh) not write-protected)
#pragma config WRT2 = OFF       
// Write Protection bit (Block 2 (004000-005FFFh) not write-protected)

#pragma config WRT3 = OFF       
// Write Protection bit (Block 3 (006000-007FFFh) not write-protected)

// CONFIG6H
#pragma config WRTC = OFF       
// Configuration Register Write Protection bit (Configuration registers 
// (300000-3000FFh) not write-protected)
#pragma config WRTB = OFF
// Boot Block Write Protection bit (Boot block (000000-0007FFh) 
// not write-protected)
#pragma config WRTD = OFF
// Data EEPROM Write Protection bit (Data EEPROM not write-protected)

// CONFIG7L
#pragma config EBTR0 = OFF      
// Table Read Protection bit (Block 0 (000800-001FFFh) not protected from 
// table reads executed in other blocks)
#pragma config EBTR1 = OFF      
// Table Read Protection bit (Block 1 (002000-003FFFh) not protected from 
// table reads executed in other blocks)
#pragma config EBTR2 = OFF      
// Table Read Protection bit (Block 2 (004000-005FFFh) not protected from table 
// reads executed in other blocks)
#pragma config EBTR3 = OFF      
// Table Read Protection bit (Block 3 (006000-007FFFh) not protected from table 
// reads executed in other blocks)

// CONFIG7H
#pragma config EBTRB = OFF      
// Boot Block Table Read Protection bit (Boot block (000000-0007FFh) 
// not protected from table reads executed in other blocks)

    /* -+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+*/
	
// Initialise starts here:
    // By default, let's declare all the unused pins as digital outputs 
// The ports:
    PORTA = 0x00;   // Reset all Flip-Flops at PORTA
    LATA = 0x00;    
    CMCON   = 0x07; // Comparator off  (RA4 and RA5 can be configured
                    // also as comparator outputs).
    ADCON1 = 0x0F;  // Configures all the PORTA and PORTB(3..0) lines as digital
	TRISA 	= 0b11000001;
    
    PORTB = 0x00; // Reset all Flip-Flops at PORTB
    LATB = 0x00; 
	TRISB 	= 0b11000001;

    PORTC = 0x00;       // Not used
    LATC = 0x00; 
    TRISC = 0b00000000; 

    PORTD = 0x00;       // Not used
    LATD = 0x00; 
    TRISD = 0x00; 
  
    PORTE = 0x00;       // Not used
    LATE = 0x00; 
    TRISE = 0b00000000; 

    
// The INT0
    INTEDG0 = 0;    // INT0 edge selection: '0' --> on falling edge 
                    //(interrupt after clicking trigger pushbutton)
    INT0IE = 1;     // Enable interrupts from the INT0 pin
	IPEN = 0; 		// Disable interrupt priority
	PEIE = 0;       // Disable peripherals interrupts 
    GIE = 1;        // Global interrupts allowed
    
	var_CLK_flag = 0;		// The flag to detect an interrupt INT
    var_current_state = State_9;  
}

//*****************************************************************************
//State variable logic routine  (The CC1 truth table and flowchart)
//*****************************************************************************
char state_logic(void){
	char error = 0;	
    switch (var_current_state) {
        case State_9:		  
            if (var_INH == 0 && var_CW == 1){	// Clockwise rotation
              var_current_state = State_5;	
            }                           // Counter-clockwise rotation
            else if (var_INH == 0 && var_CW == 0){   
              var_current_state = State_10;	
            }
            break;
        case State_5:				
            if (var_INH == 0 && var_CW == 1){		
              var_current_state = State_6;	
            }
            else if (var_INH == 0 && var_CW == 0){   
              var_current_state = State_9;	
            }
            break;
        case State_6:				
            if (var_INH == 0 && var_CW == 1){		
              var_current_state = State_10;	
            }
            else if (var_INH == 0 && var_CW == 0){   
              var_current_state = State_5;	
            }
            break;
        case State_10:				
            if (var_INH == 0 && var_CW == 1){	
              var_current_state = State_9;	
            }
            else if (var_INH == 0 && var_CW == 0){  
              var_current_state = State_6;	
            }
            break;

        default:
            error = 1;// If no states have been selected, there is an error
                        // and the systems must be stopped 
            break;
    }
	return (error);
}

//*****************************************************************************
//Output logic routine  (The CC2 truth table and flowchart)
//*****************************************************************************
char output_logic(void){
unsigned char error = 0;	 
    switch (var_current_state) { 
        case State_9:
            var_Q = 9; //or var_Q = 0b00001001;
            var_L = 9;                    
            break;
        case State_5:
            var_Q = 5;
            var_L = 5;  
            break;

        case State_6:
            var_Q = 6;
            var_L = 6;  
            break;

        case State_10:
            var_Q = 10;
            var_L = 10;  
            break;
        default:
            error = 1;	// If no states have been selected, there is
                        // an error and the systems must be stopped.
            break;
    }
	return (error);
}

// Interrupt Service Routine to attend CLK pulses to move the motor
//******************************************************************************
static void interrupt ISR(void) { //Interrupt is the key word here
    GIE = 0; //Disable interrupts while attending one of them ...
    if (INT0IF == 1) {
        var_CLK_flag = 1; //Set the variable to move the state diagram 
        // and clear the hardware interrupt flag bit (so that the system can be 
        // interrupted again when another active edge is detected)
        INT0IF = 0;
    }
    GIE = 1; 	//Enable interrupts
}


/* =============================================================================
	Poll all level-sensitive inputs and prepare convenient RAM variables  
   ========================================================================== */
static void read_inputs (void) { 
// Read all pin ports, mask and shift
	char var_buf;
    char var_buf2;
    var_buf = PORTB & 0b11000000; 
    var_buf2 = var_buf & 0b01000000; // This is INH
    var_INH = var_buf2 >>6;
    var_buf2 = var_buf & 0b10000000; // This is CW   
    var_CW = var_buf2 >>7;    
}

/* =============================================================================
	Driving output pins. Let's drive the output ports, converting 
	microcontroller RAM variables into electrical signals.    
   ========================================================================== */
void write_outputs(void){
	char var_buf;
    char var_buf2; //Local variable buffer
    
    //To write a given output pin:
    // Read and mask to preserve all the bits not to be written, and clear the 
    // bits to be written (L3, L2, L1) 
        var_buf = PORTA & 0b11110001;
        // clear and shift the bits to write 
        var_buf2 = (var_L & 0b00000111)<<1;      
        PORTA = var_buf |var_buf2; // Write all port pins in a single operation

    // Read and mask to preserve all the bits not to be written, and clear the 
    // bits to be written (L4) 
        var_buf = PORTB & 0b11011111;
        // clear and shift the bits to write 
        var_buf2 = (var_L & 0b00001000)<<2;      
        PORTB = var_buf |var_buf2; // Write all port pins in a single operation        

    // Read and mask to preserve all the bits not to be written, and clear the 
    // bits to be written (Q(4..1) 
        var_buf = PORTB & 0b11100001;
        // clear and shift the bits to write 
        var_buf2 = var_Q <<1;      
        PORTB = var_buf |var_buf2; // Write all port pins in a single operation        

}



